#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#define CHECK_OVERFLOWS 1	/* Crop? */

static int NEcho=0;
static short MasterVolume = 256;  /* 100% */

static unsigned long *EchoLen;
static unsigned short *EchoVol;
static unsigned short *EchoHead, *EchoTail;
static short **EchoBuf;

static int Heelp = 0;

static struct Option
{
    int ID;
    char c, c2;
    const char *txt, *descr;
} Options[] =
{
	/* ID-----c--c2----txt--------descr---------------- */
	{'?',   '?',  0,"help",      "Help"},
	{'V',   'V',  0,"version",   "Version information"},
	
	{0,0,0,NULL,NULL}
};

static const char ProgName[] = "ea";
static void PrintVersionInfo(void)
{
	printf(
		"Echo Amplifier (%s) version "VERSION"\n"
		"Copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)\n", ProgName);
}
static void PrintDescription(void)
{
	printf("\nThis program generates echo.\n");
}
static void SuggestHelp(void)
{
    printf("\nTry `%s --help' for more information.\n", ProgName);
}
static void ArgError(int a, const char *b)
{
    switch(a)
    {
        case 0:
            printf("%s: illegal option -- %c", ProgName, *b);
            for(a=0; Options[a].c; a++)
                if(*b == Options[a].c)
                {
                    putchar(b[1]);
                    break;
                }
            break;
        case 1:
            printf("%s: unrecognized option `--%s'", ProgName, b);
            break;
        default:
            printf("%s: invalid parameter: `%s'\n", ProgName, b);
    }
    SuggestHelp();
    exit(EXIT_FAILURE);
}
static void Case(int ID, const char **dummy)
{
	switch(ID)
	{
		case 0:	/* startup */
			break;
			
		case 'V':
			PrintVersionInfo();
			exit(EXIT_SUCCESS);
			
		case '?': Heelp  = 1; break;
		case -1: /* finish args, set default values if necessary */
			break;
	}
	/* Ensure no warnings about unused args */
	dummy = dummy;
}
static void ReadArgs(int argc, const char *const *argv,
              const char *UsageExtra, const char *UsageEnd)
{
	int a, ap;
	
	Case(0, NULL);

	for(a=1; a<argc; a++)
	{
		const char *s = argv[a];

ReArg:	if(!s)continue;
		while(*s==' ' || *s=='\t')s++;
		if(!*s)continue;

		if(*s == '-' || (*s=='/'&&s[1]=='?'))
		{
			if(*++s == '-')
			{
				int b=0;
				for(s++; Options[b].c; b++)
					if(!strcmp(s, Options[b].txt))
					{
						s = argv[++a];
						Case(Options[b].ID, &s);
						goto ReArg;
					}
				ArgError(1, s);
			}
			while(s && *s)
			{
				int b,c;
				for(b=c=0; Options[b].c; b++)
					if(s[0]==Options[b].c
					&&(s[1]==Options[b].c2
						  ||!Options[b].c2))
					{
						c=1;
						s++;
						if(Options[b].c2)s++;
						Case(Options[b].ID, &s);
						s--;
					}
				if(!c)ArgError(0, s);
				if(s)s++;
			}
		}
		else
			NEcho++;
	}
	
	#define allo(dest, type) \
		dest = (type *)malloc(NEcho * (sizeof(type))); \
		if(!dest){fprintf(stderr, "Can't allocate %d bytes for %s", \
			(int)(NEcho*(sizeof(type))), #dest); exit(errno);}
	
	allo(EchoLen, unsigned long);
	allo(EchoVol, unsigned short);
	allo(EchoHead,unsigned short);
	allo(EchoTail,unsigned short);
	allo(EchoBuf, short *);
	
	#undef allo
	
	ap=0;

	for(a=1; a<argc; a++)
	{
		const char *s = argv[a];
      	if(!s)continue;
		while(*s==' ' || *s=='\t')s++;
		if(!*s)continue;

		if(!(*s == '-' || (*s=='/'&&s[1]=='?')))
		{
			unsigned long len;
			unsigned int vol;
			
			if(!strchr(s, ':') && !ap)
			{
				MasterVolume = atoi(s)*256 / 100;
				NEcho--;
				continue;
			}
			if(sscanf(s, "%lu:%u", &len, &vol) != 2)
			{
				fprintf(stderr, "%s: Invalid argument: `%s'\n", ProgName, s);
				exit(EXIT_FAILURE);
			}
			
			EchoLen[ap] = len;
			EchoVol[ap] = vol;

			EchoHead[ap] = 0;
			EchoTail[ap] = 1;
			
			EchoBuf[ap] = (short *)calloc(EchoLen[ap], sizeof(short));
			if(!EchoBuf)
			{
				fprintf(stderr, "Can't allocate echo buffers.\n");
				exit(errno);
			}
			ap++;
		}
	}

	Case(-1, NULL);
}

int main(int argc, const char *const *argv)
{
	ReadArgs(argc, argv,
		" [globalvolume%%] {delay:volume%%}",
		"\n"
		"Examples:\n"
		"  mikmod -t96 -mird2 -e85 /mod/nes/cv2e.s3m|ea 50 1000:60 7000:10 20000:10|esdcat -m -s localhost\n"
		"  erec -2 -r44100 none|ea 10000:40 2000:20 40000:10|esdcat -s proteus\n");
	
	fprintf(stderr,
		"%s: Configured to use %d echo%s%s",
		ProgName,
		NEcho,
		NEcho-1?"es":"",
		NEcho?" (":".");
		
	{long e=0;int c;
	for(c=0; c<NEcho; c++)
	{
		if(c > 0)fprintf(stderr, ", ");
		fprintf(stderr, "%ld:%d", EchoLen[c], EchoVol[c]);
		e += EchoVol[c];
	}
	if(NEcho == 0) e = 100;
	e = e*MasterVolume/256;
	fprintf(stderr,
		")\n%s: Total gain %ld%%%s\n", ProgName, e,
		e>=100 ? " which may cause undesired effects!" : "");
	}
	
	for(;;)
	{
		#define SIZE 512
		short b[SIZE];
		register long d;
		register int a;
		int c;
		
		if(!fread(&b, sizeof(b), 1, stdin))break;
		
		for(c=0; c<SIZE; c++)
		{
			d = b[c] * MasterVolume / 256;
		
			for(a=0; a<NEcho; a++)
				d = d + EchoBuf[a][EchoTail[a]] * EchoVol[a] / 100;
			
			for(a=0; a<NEcho; a++)
			{
				EchoBuf[a][EchoHead[a]] = d;
				if(++EchoTail[a] >= EchoLen[a])EchoTail[a] = 0;
				if(++EchoHead[a] >= EchoLen[a])EchoHead[a] = 0;
			}
			
			#if CHECK_OVERFLOWS
			
			if(d <-32768) d = -32768;
			if(d > 32767) d = 32767;
			
			#endif
			
			b[c] = (short)d;
		}
		
		fwrite(&b, 1, sizeof(b), stdout);
	}
	return 0;
}
